# FourierTOuNN: Length Scale Control in Topology Optimization, using Fourier Enhanced Neural Networks
# Authors : Aaditya Chandrasekhar, Krishnan Suresh
# Affliation : University of Wisconsin - Madison
# Corresponding Author : ksuresh@wisc.edu , achandrasek3@wisc.edu
# Submitted to Structural and Multidisciplinary Optimization 
# For academic purposes only

#Versions
#Numpy 1.18.1
#Pytorch 1.5.0
#scipy 1.4.1
#cvxopt 1.2.0

#%% imports
import numpy as np
import torch
import torch.optim as optim
from os import path
from FE import FE
from plotUtil import Plotter
import matplotlib.pyplot as plt
from network import TopNet
from torch.autograd import grad
plt.rcParams['figure.dpi'] = 150


#%% main TO functionalities
class TopologyOptimizer:

    def initializeFE(self, exampleName, physics, meshFile,\
                      matProp,\
                     overrideGPU = True, penal = 3,nonDesignRegion = None):
        self.exampleName = exampleName;

        self.device = self.setDevice(overrideGPU);
        self.boundaryResolution  = 3; # default value for plotting and interpreting
        self.FE = FE();
        self.FE.initializeSolver(physics, meshFile, matProp, penal);
        self.nonDesignRegion = nonDesignRegion;
        self.xy = torch.tensor(self.FE.elemCenters, requires_grad = True).float().view(-1,2).to(self.device);
        self.nonDesignIdx = self.computeNonDesignRegion(self.xy, nonDesignRegion);
        self.xyPlot = torch.tensor(self.FE.generatePoints(self.boundaryResolution, True), requires_grad = True).float().view(-1,2).to(self.device);
        self.nonDesignPlotIdx = self.computeNonDesignRegion( self.xyPlot, nonDesignRegion);
        self.Pltr = Plotter();
    
    def setDevice(self, overrideGPU):
        if(torch.cuda.is_available() and (overrideGPU == False) ):
            device = torch.device("cuda:0");
            print("GPU enabled")
        else:
            device = torch.device("cpu")
            print("Running on CPU")
        return device;
   
    def computeNonDesignRegion(self, xy, nonDesignRegion = None): # generate points in elements
        nonDesignIdx = torch.zeros((xy.shape[0]), requires_grad = False).to(self.device);
        if(nonDesignRegion is not None):
            ctr = 0;
            for i in range(xy.shape[0]):
                if( (xy[ctr,0] < nonDesignRegion['x<']) and (xy[ctr,0] > nonDesignRegion['x>']) and (xy[ctr,1] < nonDesignRegion['y<']) and (xy[ctr,1] > nonDesignRegion['y>'])):
                        nonDesignIdx[ctr] = 1;
                ctr += 1;

        return nonDesignIdx;

    def initializeOptimizer(self, numLayers, numNeuronsPerLyr,\
                            desiredVolumeFraction, localVolumeControl,\
                            fourierMap, densityProjection,maxLengthScaleControl, \
                            symXAxis = False, symYAxis = False):
        self.desiredVolumeFraction = desiredVolumeFraction;
        self.density = self.desiredVolumeFraction*np.ones((self.FE.numElems));
        self.symXAxis = symXAxis;
        self.symYAxis = symYAxis;
        self.fourierMap = fourierMap;
        self.densityProjection = densityProjection;
        if(self.fourierMap['isOn']):
            coordnMap = np.zeros((2, self.fourierMap['numTerms']));
            for i in range(coordnMap.shape[0]):
                for j in range(coordnMap.shape[1]):
                    coordnMap[i,j] = np.random.choice([-1.,1.])*np.random.uniform(1./(2*fourierMap['maxRadius']*self.FE.elemSize[i]), 1./(2*fourierMap['minRadius']*self.FE.elemSize[i]));  #

            self.coordnMap = torch.tensor(coordnMap).float().to(self.device)#  
            inputDim = 2*self.coordnMap.shape[1];
        else:
            self.coordnMap = torch.eye(2);
            inputDim = 2;
            
        self.localVolumeControl = localVolumeControl;
        if(self.localVolumeControl['isOn']):
            self.localElems = torch.tensor(self.computeLocalElements(self.FE.nelx, self.FE.nely, self.localVolumeControl['radius'], True)).float().to(self.device);
        self.maxLengthScaleControl = maxLengthScaleControl;
        if(self.maxLengthScaleControl['isOn']):
            self.localElems = torch.tensor(self.computeLocalElements(self.FE.nelx, self.FE.nely, self.localVolumeControl['radius'], False)).float().to(self.device);
        self.topNet = TopNet(numLayers, numNeuronsPerLyr, inputDim).to(self.device);
        self.objective = 0.;

#%%
    def computeLocalElements(self, nelx, nely, dist, avgLocality):
        localElems = np.zeros((nelx*nely, nelx*nely));
        for elem in range(nelx*nely):
            ex = self.FE.elemSize[0]*elem//nely;
            ey = self.FE.elemSize[1]*elem%nely;
            for neigh in range(nelx*nely):
                nx = self.FE.elemSize[0]*neigh//nely;
                ny = self.FE.elemSize[1]*neigh%nely;
                r = (nx-ex)**2 + (ny-ey)**2;
                if(r <= dist):
                    localElems[elem, neigh] = 1;
            if(avgLocality):
                localElems[elem, :] /= np.sum(localElems[elem,:]) ;
        return localElems;
    
    def computeLocalVolumeConstraint(self, localElems, rho, penal = 6.):
        localVolCons = torch.tensor([0.]).to(self.device);
        localVolCons = torch.pow(torch.mean(torch.pow(torch.matmul(localElems, rho), penal)), 1./penal)/self.desiredVolumeFraction - 1;
        return localVolCons;
    
    def computeMaxLengthScaleConstraint(self, localElems, rho, epoch, penal = -6.):
        n =  min(3., 1. + epoch*0.05);
        voidVol = torch.matmul(localElems, (1.02-rho)**n);
        minVoidVol = torch.pow(torch.sum(voidVol**penal), 1./penal);
        consVal = (minVoidVol/self.maxLengthScaleControl['voidVol']) - 1;
        return consVal;
        
    def applySymmetry(self, x):
        if(self.symYAxis['isOn']):
            xv =( self.symYAxis['midPt'] + torch.abs( x[:,0] - self.symYAxis['midPt']));
        else:
            xv = x[:,0];
        if(self.symXAxis['isOn']):
            yv = (self.symXAxis['midPt'] + torch.abs( x[:,1] - self.symXAxis['midPt'])) ;
        else:
            yv = x[:,1];
        x = torch.transpose(torch.stack((xv,yv)),0,1);
        return x;
    
    def applyFourierMapping(self, x):
        if(self.fourierMap['isOn']):
            c = torch.cos(2*np.pi*torch.matmul(x,self.coordnMap));
            s = torch.sin(2*np.pi*torch.matmul(x,self.coordnMap));
            xv = torch.cat((c,s), axis = 1);
            return xv;
        return x;
    
    def projectDensity(self, x):
        if(self.densityProjection['isOn']):
            b = self.densityProjection['sharpness']
            nmr = np.tanh(0.5*b) + torch.tanh(b*(x-0.5));
            x = 0.5*nmr/np.tanh(0.5*b);
        return x;

    def optimizeDesign(self,maxEpochs, minEpochs,useSavedNet):
        self.convergenceHistory = {'compliance':[], 'vol':[], 'grayElems':[]};
        savedNetFileName = "./results/" + self.exampleName + "_"  + str(self.FE.numElems) +  '.nt';
        learningRate = 0.01;
        alphaMax = 100*self.desiredVolumeFraction;
        alphaIncrement= 0.05;
        alpha = alphaIncrement; # start
        nrmThreshold = 0.01; # for gradient clipping
        if(useSavedNet):
            if (path.exists(savedNetFileName)):
                self.topNet = torch.load(savedNetFileName);
            else:
                print("Network file not found");
        self.optimizer = optim.Adam(self.topNet.parameters(), amsgrad=True,lr=learningRate);

        for epoch in range(maxEpochs):
            
            self.optimizer.zero_grad();
            batch_x = self.applySymmetry(self.xy);
            x = self.applyFourierMapping(batch_x);
            nn_rho = torch.flatten(self.topNet(x,self.nonDesignIdx)).to(self.device);
            nn_rho = self.projectDensity(nn_rho);
            rho_np = nn_rho.cpu().detach().numpy(); # move tensor to numpy array
            self.density = rho_np;
            u, Jelem = self.FE.solve(rho_np); # Call FE 88 line code [Niels Aage 2013]

            if(epoch == 0):
                self.obj0 = ( self.FE.Emax*(rho_np**self.FE.penal)*Jelem).sum()
            # For sensitivity analysis, exponentiate by 2p here and divide by p in the loss func hence getting -ve sign
                
            Jelem = np.array(self.FE.Emax*(rho_np**(2*self.FE.penal))*Jelem).reshape(-1);
            Jelem = torch.tensor(Jelem).view(-1).float().to(self.device) ;
            objective = torch.sum(torch.div(Jelem,nn_rho**self.FE.penal))/self.obj0; # compliance
            if(self.localVolumeControl['isOn']):
                volConstraint = self.computeLocalVolumeConstraint(self.localElems, nn_rho); 
            else:
                volConstraint =((torch.mean(nn_rho)/self.desiredVolumeFraction) - 1.0); # global vol constraint
            currentVolumeFraction = np.average(rho_np);
            self.objective = objective;
            loss =   self.objective+ alpha*torch.pow(volConstraint,2);

            if(self.maxLengthScaleControl['isOn']):
               lenScaleCons = self.computeMaxLengthScaleConstraint(self.localElems, nn_rho, epoch);
               loss += alpha*torch.pow(lenScaleCons, 2);

            alpha = min(alphaMax, alpha + alphaIncrement);
            loss.backward(retain_graph=True);
            torch.nn.utils.clip_grad_norm_(self.topNet.parameters(),nrmThreshold)
            self.optimizer.step();
            greyElements= sum(1 for rho in rho_np if ((rho > 0.2) & (rho < 0.8)));
            relGreyElements = self.desiredVolumeFraction*greyElements/len(rho_np);
            self.convergenceHistory['compliance'].append(self.objective.item());
            self.convergenceHistory['vol'].append(currentVolumeFraction);
            self.convergenceHistory['grayElems'].append(relGreyElements);
            self.FE.penal = min(8.0,self.FE.penal + 0.02); # continuation scheme
            if(epoch % 20 == 0):
                titleStr = "Iter {:d} , Obj {:.2F} , vol {:.2F}".format(epoch, self.objective.item()*self.obj0, currentVolumeFraction);
                figName = "./frames/"+self.exampleName+"_"+str(epoch)+".png";
                self.Pltr.plotDensity(self.xy.detach().cpu().numpy(), rho_np.reshape((self.FE.nelx, self.FE.nely)), titleStr, saveFig = {'isOn':True, 'name':figName}, retainPlot = True);
                print("{:3d} J: {:.2F}; Vf: {:.3F}; loss: {:.3F}; relGreyElems: {:.3F} "\
                  .format(epoch, self.objective.item()*self.obj0 ,currentVolumeFraction,loss.item(),relGreyElements));
            if ((epoch > minEpochs ) & (relGreyElements < 0.025)):
                break;
        figName = "./data/"+self.exampleName+"_"+"final.png";
        self.Pltr.plotDensity(self.xy.detach().cpu().numpy(), rho_np.reshape((self.FE.nelx, self.FE.nely)), titleStr, saveFig = {'isOn':True, 'name':figName}, retainPlot = True);
        print("{:3d} J: {:.2F}; Vf: {:.3F}; loss: {:.3F}; relGreyElems: {:.3F} "\
             .format(epoch, self.objective.item()*self.obj0 ,currentVolumeFraction,loss.item(),relGreyElements)); 

        print("Final J : {:.3f}".format(self.objective.item()*self.obj0));
        torch.save(self.topNet, savedNetFileName);
        self.Pltr.plotConvergence(self.convergenceHistory, sameFig = False);
        
        batch_x = self.applySymmetry(self.xyPlot);
        x = self.applyFourierMapping(batch_x);
        rho = torch.flatten(self.projectDensity(self.topNet(x,self.nonDesignPlotIdx)));
        rho_np = rho.cpu().detach().numpy();
        
        titleStr = "Iter {:d} , Obj {:.2F} , vol {:.2F}".format(epoch, self.objective.item()*self.obj0, currentVolumeFraction);
        figName = "./data/"+self.exampleName+"_"+"highRes.png";
        self.Pltr.plotDensity(self.xyPlot.detach().cpu().numpy(), rho_np.reshape((self.FE.nelx*self.boundaryResolution, self.FE.nely*self.boundaryResolution)), titleStr, saveFig = {'isOn':True, 'name':figName}, retainPlot = False);
        pltGradient= False
        if(pltGradient):
            z = torch.ones(rho.shape).to(self.device) # temp var
            drho_dx = grad(rho, self.xyPlot, grad_outputs=z, create_graph=True)[0].cpu().detach().numpy();
            norm_drho_dx = np.sqrt(drho_dx[:,0]**2 + drho_dx[:,1]**2).reshape((self.boundaryResolution*self.FE.nelx, self.boundaryResolution*self.FE.nely));
            self.Pltr.plotDensity(self.xy.detach().cpu().numpy(), np.log(norm_drho_dx), titleStr, saveFig = {'isOn':True, 'name':figName}, retainPlot = False);
        return self.convergenceHistory;

